from numpy import flipud, meshgrid, ones, zeros
from numpy.linalg import norm, solve
from numba import jit
from sys import argv
from pickle import dump, load
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve, lgmres
from scipy.stats import gmean, hmean

import numpy as np
import numba as nb
import pylab

poro = 0.25

stem = argv[1]
pickle_jar_name = 'k_field_' + stem + '.pickle'
with open(pickle_jar_name, 'rb') as pickle_file:
    dump_data = load(pickle_file)
width_x = np.float64(dump_data['width_x'])
width_y = np.float64(dump_data['width_y'])

K = np.float64(dump_data['field'])

size_x = np.int32(K.shape[1])
size_y = np.int32(K.shape[0])

dx = np.float64(width_x/size_x)
dy = np.float64(width_y/size_y)
h_max  = np.float64(1.)
h_min = np.float64(0.)

L = zeros([size_y, size_x], dtype=np.float64)
M = zeros([size_y, size_x], dtype=np.float64)

Qx = zeros([size_y, size_x], dtype=np.float64)
Qy = zeros([size_y, size_x], dtype=np.float64)

b = zeros(size_x*size_y, dtype=np.float64)

@jit(nb.types.UniTuple(nb.int32,2)(nb.int32))
def coord_indexes(line_idx):
    x_idx = line_idx % size_x
    y_idx = int(line_idx / size_x)
    return (y_idx, x_idx)

@jit(nb.int32(nb.int32,nb.int32))
def flat_index(y_idx, x_idx): 
    line_idx = x_idx + y_idx*size_x
    return (line_idx)

@jit
def populate_LM(K, L, M, size_x, size_y):
    for x_idx in range(1, size_x-1):
        for y_idx in range(1, size_y-1):
            L[y_idx, x_idx] = (K[y_idx, x_idx + 1] - K[y_idx, x_idx - 1])/(2*dx)
            M[y_idx, x_idx] = (K[y_idx + 1, x_idx] - K[y_idx - 1, x_idx])/(2*dy)

@jit(forceobj=True)
def populate_Ab(b, K, L, M, size_x, size_y):
    rows = []
    cols = []
    data = []
    for row_idx in range(size_x*size_y):
        y_idx, x_idx = coord_indexes(row_idx)
        if y_idx == 0:
            rows.append(row_idx)
            cols.append(row_idx)
            data.append(1)
            b[row_idx] = h_min

        elif y_idx == size_y - 1:
            rows.append(row_idx)
            cols.append(row_idx)
            data.append(1)
            b[row_idx] = h_max

        elif x_idx == 0:
            rows.extend((row_idx,row_idx,row_idx))
            cols.extend((row_idx,row_idx+1,row_idx+2))
            data.extend((1,-4./3.,1./3))

        elif x_idx == size_x - 1:
            rows.extend((row_idx,row_idx,row_idx))
            cols.extend((row_idx,row_idx-1,row_idx-2))
            data.extend((1,-4./3.,1./3))

        else:
            rows.extend((row_idx,row_idx,row_idx,row_idx,row_idx))
            cols.extend((
                flat_index(y_idx, x_idx+1), 
                flat_index(y_idx, x_idx-1), 
                flat_index(y_idx+1, x_idx),
                flat_index(y_idx-1, x_idx),
                row_idx
                ))
            data.extend((
                K[y_idx, x_idx]/dx**2 + L[y_idx, x_idx]/(2*dx),
                K[y_idx, x_idx]/dx**2 - L[y_idx, x_idx]/(2*dx),
                K[y_idx, x_idx]/dy**2 + M[y_idx, x_idx]/(2*dy),
                K[y_idx, x_idx]/dy**2 - M[y_idx, x_idx]/(2*dy),
                - 2*K[y_idx,x_idx]/dx**2 - 2*K[y_idx,x_idx]/dy**2 
                ))
    A = csc_matrix((data,(rows,cols)),shape=(size_y*size_x,size_y*size_x),dtype=np.float64)
    return(A)

populate_LM(K, L, M, size_x, size_y)
A = populate_Ab(b, K, L, M, size_x, size_y)

h = spsolve(A,b)
# tmp,Y = meshgrid(np.linspace(0,1,size_x),np.linspace(h_min,h_max,size_y))
# h0 = Y.flatten()
# h, exit_code = lgmres(A,b, x0=h0, maxiter=5000)
# print("Exit code", exit_code)

quality = norm(A.dot(h) - b)
print("Solution error", quality)

H = h.reshape([size_y,size_x])

Q_out = sum((H[size_y-1,:]-H[size_y-2,:])/dy* 2*(K[size_y-1,:]**-1+K[size_y-2,:]**-1)**-1 *dx)
Q_in = sum((H[1,:]-H[0,:])/dy* 2*(K[0,:]**-1+K[1,:]**-1)**-1 *dx)
print(Q_out,Q_in)
Q_global = 10**np.mean(np.log10(K))*(h_max-h_min)/(size_y*dy)*(size_x*dx)
print(Q_global)

@jit
def populate_QxQy(Qx, Qy, H, K, dx, dy, size_x, size_y):
    for x_idx in range(1, size_x-1):
        for y_idx in range(0, size_y):
            k_harm = (0.25*K[y_idx, x_idx-1]**-1 + 0.5*K[y_idx, x_idx]**-1 + 0.25*K[y_idx, x_idx+1]**-1)**-1
            Qx[y_idx, x_idx] = -(H[y_idx, x_idx + 1] - H[y_idx, x_idx - 1])/(2*dx)*k_harm
    
    for x_idx in range(0, size_x):
        for y_idx in range(1, size_y-1):
            k_harm = (0.25*K[y_idx-1, x_idx]**-1 + 0.5*K[y_idx, x_idx]**-1 + 0.25*K[y_idx+1, x_idx]**-1)**-1
            Qy[y_idx, x_idx] = -(H[y_idx + 1, x_idx] - H[y_idx - 1, x_idx])/(2*dy)*K[y_idx, x_idx]
    
    k_harm_vec = (0.5*K[size_y-1,:]**-1 + 0.5*K[size_y-2,:]**-1)**-1
    Qy[size_y-1,:] = -(H[size_y-1,:]-H[size_y-2,:])/dy* k_harm_vec

    k_harm_vec = (0.5*K[1,:]**-1 + 0.5*K[0,:]**-1)**-1
    Qy[0,:] = -(H[1,:]-H[0,:])/dy*k_harm_vec

populate_QxQy(Qx, Qy, H, K, dx, dy, size_x, size_y)

speeds = np.sqrt(Qx[1:(size_y-1),1:(size_x-1)]**2+Qy[1:(size_y-1),1:(size_x-1)]**2)
print("Mean velocities (A,G,H)", np.mean(speeds),gmean(speeds, axis=None),hmean(speeds, axis=None))

#DRAW QUIVER PLOT
xaxis = np.linspace(dx/2,width_x-dx/2,size_x)
yaxis = np.linspace(dy/2,width_y-dy/2,size_y)
X,Y = np.meshgrid(xaxis,yaxis)

fig = pylab.figure()
ax = fig.add_subplot(111)
pylab.imshow(flipud(H),extent=[0,width_x,0,width_y], alpha=0.5)
qu = pylab.quiver(X,Y,Qx,Qy, scale=1e-1)
qu.set_facecolor('k')
ax.set_aspect('equal')

pylab.xlabel('x [m]')
pylab.ylabel('y [m]')
pylab.show()

#PICKLE
xaxis = np.linspace(dx/2, width_x - dx/2, size_x)
yaxis = np.linspace(dy/2, width_y - dy/2, size_y)

pickle_filename = 'v_field_' + stem +'.pickle'
pickle_dict = {
    'xaxis' : xaxis,
    'yaxis' : yaxis,
    'vX': Qx/poro,
    'vY': Qy/poro,
    'width_x': width_x,
    'width_y': width_y}
with open(pickle_filename, 'wb') as pickle_jar:
    dump(pickle_dict, pickle_jar)
